#define WIN32_LEAN_AND_MEAN
#define WINVER 0x501
#define _WIN32_WINNT 0x501


#include <windows.h>
#include <psapi.h>
#include <stdio.h>

struct cproc
{
	DWORD pid;
	HANDLE h;
	struct cproc *l;
	struct cproc *r;
};

struct cproc rootproc;
struct cproc *childprocs = &rootproc;

HANDLE hh;//heap handle
DWORD winerrno;

static inline void printusage()
{
	_putws(L"win-wait-for-all-children.exe SOME_PROGRAM_TO_RUN.exe [ARGUMENTS_FOR_PROGRAM] [...]");
	return;
}

__attribute__((noreturn)) static inline void myexit(DWORD x)
{
	wprintf(L"win-wait-for-all-children.exe: Exiting with exit code 0x%x\n", x);
	ExitProcess(x);
	//__builtin_unreachable();
}

static inline void newproc(DWORD pid, HANDLE h)
{
	struct cproc *curpos;
	struct cproc *n;

	n = HeapAlloc(hh, 0, sizeof(struct cproc));
	if (!n)
	{
		CloseHandle(h);
		_putws(L"\t\tHeapAlloc() failed!");
		return;
	}

	n->pid = pid;
	n->h = h;

	n->l = NULL;
	n->r = NULL;

	curpos = childprocs;
	while (1)
	{
		if (pid < curpos->pid)
		{
			if (!curpos->l)
			{
				curpos->l = n;
				return;
			}
			curpos = curpos->l;
			continue;
		}
		if (pid > curpos->pid)
		{
			if (!curpos->r)
			{
				curpos->r = n;
				return;
			}
			curpos = curpos->r;
			continue;
		}
		if (curpos != &rootproc)
		{
			curpos->h = h;
		}
		else
		{
			CloseHandle(h);
		}
		HeapFree(hh, 0, n);
		return;
	}
}

static inline struct cproc *findproc(DWORD pid)
{
	struct cproc *curpos;

	curpos = childprocs;
	while (curpos)
	{
		if (pid < curpos->pid)
		{
			curpos = curpos->l;
			continue;
		}
		if (pid > curpos->pid)
		{
			curpos = curpos->r;
			continue;
		}
		break;
	}
	return curpos;
}

static inline DWORD runexe(wchar_t *exe)
{
	STARTUPINFO sui = {0};
	PROCESS_INFORMATION pi = {0};
	DWORD ecode;
	HANDLE job;
	HANDLE iocompport;
	JOBOBJECT_ASSOCIATE_COMPLETION_PORT jobcp;
	DWORD jobcompcode;
	ULONG_PTR jobcompkey;
	LPOVERLAPPED jobcompoverlap;
	HANDLE cph;//child process handle
	wchar_t cppath[MAX_PATH];
	struct cproc *curpos;

	sui.cb = sizeof(STARTUPINFO);

	//
	job = CreateJobObject(NULL, NULL);
	if (!job)
	{
		winerrno = GetLastError();
		_putws(L"win-wait-for-all-children.exe: CreateJobObject() failed");
		myexit(winerrno);
	}

	iocompport = CreateIoCompletionPort(INVALID_HANDLE_VALUE, NULL, 0, 1);
	if (!iocompport)
	{
		winerrno = GetLastError();
		wprintf(L"win-wait-for-all-children.exe: CreateIoCompletionPort() failed errno 0x%x\n", winerrno);
		myexit(winerrno);
	}

	jobcp.CompletionKey = job;
	jobcp.CompletionPort = iocompport;
	if (!SetInformationJobObject(job, JobObjectAssociateCompletionPortInformation, &jobcp, sizeof(JOBOBJECT_ASSOCIATE_COMPLETION_PORT)))
	{
		winerrno = GetLastError();
		wprintf(L"win-wait-for-all-children.exe: SetInformationJobObject() failed errno 0x%x\n", winerrno);
		myexit(winerrno);
	}

	if (!CreateProcess(NULL, exe, NULL, NULL, 0, CREATE_SUSPENDED, NULL, NULL, &sui, &pi))
	{
		winerrno = GetLastError();
		wprintf(L"win-wait-for-all-children.exe: CreateProcess() failed errno 0x%x\n", winerrno);
		myexit(winerrno);
	}
	wprintf(L"win-wait-for-all-children.exe: CreateProcess() made pid = %u\n", pi.dwProcessId);

	rootproc.pid = pi.dwProcessId;
	rootproc.h = pi.hProcess;

	if (!AssignProcessToJobObject(job, pi.hProcess))
	{
		winerrno = GetLastError();
		wprintf(L"win-wait-for-all-children.exe: AssignProcessToJobObject() failed errno 0x%x\n", winerrno);
		_putws(L"Terminating the suspended child...");
		if (!TerminateProcess(pi.hProcess, 0))
		{
			winerrno = GetLastError();
			wprintf(L"win-wait-for-all-children.exe: TerminateProcess() failed errno 0x%x\n", winerrno);
		}
		myexit(winerrno);
	}

	if (ResumeThread(pi.hThread) == ((DWORD)-1))
	{
		winerrno = GetLastError();
		wprintf(L"win-wait-for-all-children.exe: ResumeThread() failed errno 0x%x\n", winerrno);
		if (!TerminateProcess(pi.hProcess, 0))
		{
			winerrno = GetLastError();
			wprintf(L"win-wait-for-all-children.exe: TerminateProcess() failed errno 0x%x\n", winerrno);
		}
		myexit(winerrno);
	}

	do
	{
		GetQueuedCompletionStatus(iocompport, &jobcompcode, &jobcompkey, &jobcompoverlap, INFINITE);
		if (jobcompcode == JOB_OBJECT_MSG_NEW_PROCESS)
		{
			cph = OpenProcess(PROCESS_QUERY_INFORMATION, 0, (DWORD)jobcompoverlap);
			if (!cph)
			{
				winerrno = GetLastError();
				wprintf(L"OpenProcess() failed errno 0x%x\n", winerrno);
			}
			else
			{
				if (!GetProcessImageFileName(cph, cppath, MAX_PATH))
				{
					winerrno = GetLastError();
					wprintf(L"GetProcessImageFileName() failed errno 0x%x\n", winerrno);
				}
				else
				{
					wprintf(L"file \"%ls\"\n", cppath);
				}
				newproc((DWORD)jobcompoverlap, cph);
			}
			wprintf(L"\tnew process pid %u\n", jobcompoverlap);
		}
		else if (jobcompcode == JOB_OBJECT_MSG_EXIT_PROCESS)
		{
			curpos = findproc((DWORD)jobcompoverlap);
			if (curpos)
			{
				if (!GetExitCodeProcess(curpos->h, &ecode))
				{
					winerrno = GetLastError();
					wprintf(L"win-wait-for-all-children.exe: GetExitCodeProcess() failed errno 0x%x\n", winerrno);
				}
				wprintf(L"\texit process pid %u ecode 0x%x\n", jobcompoverlap, ecode);
			}
			else
			{
				wprintf(L"\texit process pid %u\n", jobcompoverlap);
			}
		}
		else if (jobcompcode == JOB_OBJECT_MSG_ABNORMAL_EXIT_PROCESS)
		{
			curpos = findproc((DWORD)jobcompoverlap);
			if (curpos)
			{
				if (!GetExitCodeProcess(curpos->h, &ecode))
				{
					winerrno = GetLastError();
					wprintf(L"win-wait-for-all-children.exe: GetExitCodeProcess() failed errno 0x%x\n", winerrno);
				}
				wprintf(L"\tcrash process pid %u ecode 0x%x\n", jobcompoverlap, ecode);
			}
			else
			{
				wprintf(L"\tcrash process pid %u\n", jobcompoverlap);
			}
		}
	}
	while ((((HANDLE)jobcompkey) != job) || (jobcompcode != JOB_OBJECT_MSG_ACTIVE_PROCESS_ZERO));

	if (!GetExitCodeProcess(pi.hProcess, &ecode))
	{
		winerrno = GetLastError();
		wprintf(L"win-wait-for-all-children.exe: GetExitCodeProcess() failed errno 0x%x\n", winerrno);
		myexit(winerrno);
	}
	CloseHandle(iocompport);
	CloseHandle(job);
	CloseHandle(pi.hThread);
	CloseHandle(pi.hProcess);
	wprintf(L"root process exited with exit code 0x%x\n", ecode);
	return ecode;
}

int wmain(int argc, wchar_t *argv[], wchar_t *envp[])
{
	int curarg;
	unsigned curslen;
	wchar_t *argstring;
	unsigned argstringsize=0;
	/* ULONGLONG uptimems; */
	/* double uptimehours; */

	(void)envp;

	if (argc<2)
	{
		printusage();
		return ERROR_INVALID_PARAMETER;
	}

	hh = HeapCreate(HEAP_NO_SERIALIZE, 0, 0);
	if (!hh)
	{
		winerrno = GetLastError();
		_putws(L"win-wait-for-all-children.exe: HeapCreate() failed!");
		ExitProcess(winerrno);
	}

	//Because I have unicode enabled I need to have the strings in a
	//modifiable area
	curarg = 1;
	while (curarg != argc)
	{
		argstringsize += wcslen(argv[curarg]);
		curarg++;
	}
	//add room for spaces and nulls
	argstringsize += curarg;

	argstring = HeapAlloc(hh, 0, argstringsize * sizeof(wchar_t));
	if (!argstring)
	{
		_putws(L"win-wait-for-all-children.exe: HeapAlloc() failed!");
		myexit(ERROR_NOT_ENOUGH_MEMORY);
	}

	curarg = 1;
	argstringsize = 0;
	while (curarg != argc)
	{
		curslen = wcslen(argv[curarg]);
		__builtin_memcpy(&argstring[argstringsize], argv[curarg], curslen * sizeof(wchar_t));
		argstringsize += 1 + curslen;
		argstring[argstringsize-1] = L' ';
		curarg++;
	}
	argstring[argstringsize] = 0;//null terminator

	winerrno = runexe(argstring);
	myexit(winerrno);

	return 0;
}



